@@ -43,10 +43,10 @@ import           Control.Arrow
 import           Control.Monad                         ( join, MonadPlus(..) )
 
 import qualified Data.Foldable                    as F
+import           Data.Monoid
 
 import           Text.Pandoc.Readers.Odt.Generic.Fallible
 import           Text.Pandoc.Readers.Odt.Generic.Utils
-import           Text.Pandoc.Compat.Monoid
 
 and2 :: (Arrow a) => a b c -> a b c' -> a b (c,c')
 and2 = (&&&)
@@ -383,70 +383,70 @@ raiseAEmpty       = arr (fromRight (const mempty) >>> Left)
 
 
 -- | Execute the second arrow if the first succeeds
-(>>?) :: (ArrowChoice a, Monoid failure)
+(>>?) :: (ArrowChoice a)
             => FallibleArrow a x       failure success
             -> FallibleArrow a success failure success'
             -> FallibleArrow a x       failure success'
 a >>? b = a >>> Left ^||| b
 
 -- | Execute the lifted second arrow if the first succeeds
-(>>?^) :: (ArrowChoice a, Monoid failure)
+(>>?^) :: (ArrowChoice a)
             => FallibleArrow a x       failure success
             -> (success                     -> success')
             -> FallibleArrow a x       failure success'
 a >>?^ f = a >>^ Left ^|||^ Right . f
 
 -- | Execute the lifted second arrow if the first succeeds
-(>>?^?) :: (ArrowChoice a, Monoid failure)
+(>>?^?) :: (ArrowChoice a)
             => FallibleArrow a x       failure success
             -> (success      -> Either failure success')
             -> FallibleArrow a x       failure success'
 a >>?^? b = a >>> Left ^|||^ b
 
 -- | Execute the second arrow if the lifted first arrow succeeds
-(^>>?) :: (ArrowChoice a, Monoid failure)
+(^>>?) :: (ArrowChoice a)
             => (x            -> Either failure success)
             -> FallibleArrow a success failure success'
             -> FallibleArrow a x       failure success'
 a ^>>? b = a ^>> Left ^||| b
 
 -- | Execute the lifted second arrow if the lifted first arrow succeeds
-(^>>?^) :: (ArrowChoice a, Monoid failure)
+(^>>?^) :: (ArrowChoice a)
             => (x            -> Either failure success)
             -> (success                     -> success')
             -> FallibleArrow a x       failure success'
 a ^>>?^ f = arr $ a >>> right f
 
 -- | Execute the lifted second arrow if the lifted first arrow succeeds
-(^>>?^?) :: (ArrowChoice a, Monoid failure)
+(^>>?^?) :: (ArrowChoice a)
             => (x            -> Either failure success)
             -> (success      -> Either failure success')
             -> FallibleArrow a x       failure success'
 a ^>>?^? f = a ^>> Left ^|||^ f
 
 -- | Execute the second, non-fallible arrow if the first arrow succeeds
-(>>?!) :: (ArrowChoice a, Monoid failure)
+(>>?!) :: (ArrowChoice a)
             => FallibleArrow a x       failure success
             ->               a success         success'
             -> FallibleArrow a x       failure success'
 a >>?! f = a >>> right f
 
 ---
-(>>?%) :: (ArrowChoice a, Monoid f)
+(>>?%) :: (ArrowChoice a)
           => FallibleArrow a x f (b,b')
           -> (b -> b' -> c)
           -> FallibleArrow a x f c
 a >>?% f = a >>?^ (uncurry f)
 
 ---
-(^>>?%) :: (ArrowChoice a, Monoid f)
+(^>>?%) :: (ArrowChoice a)
           => (x -> Either f (b,b'))
           -> (b -> b' -> c)
           -> FallibleArrow a x f c
 a ^>>?% f = arr a >>?^ (uncurry f)
 
 ---
-(>>?%?) :: (ArrowChoice a, Monoid f)
+(>>?%?) :: (ArrowChoice a)
            => FallibleArrow a x f (b,b')
            -> (b -> b' -> (Either f c))
            -> FallibleArrow a x f c
