@@ -2,7 +2,7 @@
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE ViewPatterns #-}
 {-
-Copyright (C) 2014-2016 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
+Copyright (C) 2014-2017 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 {- |
    Module      : Text.Pandoc.Readers.Org.Options
-   Copyright   : Copyright (C) 2014-2016 Albert Krewinkel
+   Copyright   : Copyright (C) 2014-2017 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -34,27 +34,26 @@ module Text.Pandoc.Readers.Org.Blocks
   ) where
 
 import           Text.Pandoc.Readers.Org.BlockStarts
-import           Text.Pandoc.Readers.Org.ExportSettings ( exportSettings )
 import           Text.Pandoc.Readers.Org.Inlines
+import           Text.Pandoc.Readers.Org.Meta ( metaExport, metaKey, metaLine )
 import           Text.Pandoc.Readers.Org.ParserState
 import           Text.Pandoc.Readers.Org.Parsing
 import           Text.Pandoc.Readers.Org.Shared
-                   ( isImageFilename, rundocBlockClass, toRundocAttrib
-                   , translateLang )
+                   ( cleanLinkString, isImageFilename, rundocBlockClass
+                   , toRundocAttrib, translateLang )
 
 import qualified Text.Pandoc.Builder as B
 import           Text.Pandoc.Builder ( Inlines, Blocks )
 import           Text.Pandoc.Definition
-import           Text.Pandoc.Compat.Monoid ((<>))
 import           Text.Pandoc.Options
-import           Text.Pandoc.Shared ( compactify', compactify'DL )
+import           Text.Pandoc.Shared ( compactify', compactify'DL, safeRead )
 
 import           Control.Monad ( foldM, guard, mzero, void )
 import           Data.Char ( isSpace, toLower, toUpper)
-import           Data.List ( foldl', intersperse, isPrefixOf )
-import qualified Data.Map as M
+import           Data.Default ( Default )
+import           Data.List ( foldl', isPrefixOf )
 import           Data.Maybe ( fromMaybe, isNothing )
-import           Network.HTTP ( urlEncode )
+import           Data.Monoid ((<>))
 
 --
 -- Org headers
@@ -82,12 +81,17 @@ newtype PropertyValue = PropertyValue { fromValue :: String }
 toPropertyValue :: String -> PropertyValue
 toPropertyValue = PropertyValue
 
+-- | Check whether the property value is non-nil (i.e. truish).
+isNonNil :: PropertyValue -> Bool
+isNonNil p = map toLower (fromValue p) `notElem` ["()", "{}", "nil"]
+
 -- | Key/value pairs from a PROPERTIES drawer
 type Properties = [(PropertyKey, PropertyValue)]
 
 -- | Org mode headline (i.e. a document subtree).
 data Headline = Headline
   { headlineLevel      :: Int
+  , headlineTodoMarker :: Maybe TodoMarker
   , headlineText       :: Inlines
   , headlineTags       :: [Tag]
   , headlineProperties :: Properties
@@ -105,6 +109,7 @@ headline :: Int -> OrgParser (F Headline)
 headline lvl = try $ do
   level <- headerStart
   guard (lvl <= level)
+  todoKw <- optionMaybe todoKeyword
   title <- trimInlinesF . mconcat <$> manyTill inline endOfTitle
   tags  <- option [] headerTags
   newline
@@ -117,6 +122,7 @@ headline lvl = try $ do
     children' <- sequence children
     return $ Headline
       { headlineLevel = level
+      , headlineTodoMarker = todoKw
       , headlineText = title'
       , headlineTags = tags
       , headlineProperties = properties
@@ -191,23 +197,46 @@ headlineToHeaderWithContents hdln@(Headline {..}) = do
 
 headlineToHeader :: Headline -> OrgParser Blocks
 headlineToHeader (Headline {..}) = do
-  let text        = tagTitle headlineText headlineTags
+  exportTodoKeyword <- getExportSetting exportWithTodoKeywords
+  let todoText    = if exportTodoKeyword
+                    then case headlineTodoMarker of
+                      Just kw -> todoKeywordToInlines kw <> B.space
+                      Nothing -> mempty
+                    else mempty
+  let text        = tagTitle (todoText <> headlineText) headlineTags
   let propAttr    = propertiesToAttr headlineProperties
   attr           <- registerHeader propAttr headlineText
   return $ B.headerWith attr headlineLevel text
 
+todoKeyword :: OrgParser TodoMarker
+todoKeyword = try $ do
+  taskStates <- activeTodoMarkers <$> getState
+  let kwParser tdm = try $ (tdm <$ string (todoMarkerName tdm) <* spaceChar)
+  choice (map kwParser taskStates)
+
+todoKeywordToInlines :: TodoMarker -> Inlines
+todoKeywordToInlines tdm =
+  let todoText  = todoMarkerName tdm
+      todoState = map toLower . show $ todoMarkerState tdm
+      classes = [todoState, todoText]
+  in B.spanWith (mempty, classes, mempty) (B.str todoText)
+
 propertiesToAttr :: Properties -> Attr
 propertiesToAttr properties =
   let
     toStringPair prop = (fromKey (fst prop), fromValue (snd prop))
     customIdKey = toPropertyKey "custom_id"
     classKey    = toPropertyKey "class"
+    unnumberedKey = toPropertyKey "unnumbered"
+    specialProperties = [customIdKey, classKey, unnumberedKey]
     id'  = fromMaybe mempty . fmap fromValue . lookup customIdKey $ properties
     cls  = fromMaybe mempty . fmap fromValue . lookup classKey    $ properties
-    kvs' = map toStringPair . filter ((`notElem` [customIdKey, classKey]) . fst)
+    kvs' = map toStringPair . filter ((`notElem` specialProperties) . fst)
            $ properties
+    isUnnumbered =
+      fromMaybe False . fmap isNonNil . lookup unnumberedKey $ properties
   in
-    (id', words cls, kvs')
+    (id', words cls ++ (if isUnnumbered then ["unnumbered"] else []), kvs')
 
 tagTitle :: Inlines -> [Tag] -> Inlines
 tagTitle title tags = title <> (mconcat $ map tagToInline tags)
@@ -232,8 +261,8 @@ blockList = do
 -- | Get the meta information safed in the state.
 meta :: OrgParser Meta
 meta = do
-  st <- getState
-  return $ runF (orgStateMeta st) st
+  meta' <- metaExport
+  runF meta' <$> getState
 
 blocks :: OrgParser (F Blocks)
 blocks = mconcat <$> manyTill block (void (lookAhead headerStart) <|> eof)
@@ -266,6 +295,17 @@ data BlockAttributes = BlockAttributes
   , blockAttrKeyValues :: [(String, String)]
   }
 
+-- | Convert BlockAttributes into pandoc Attr
+attrFromBlockAttributes :: BlockAttributes -> Attr
+attrFromBlockAttributes (BlockAttributes{..}) =
+  let
+    ident   = fromMaybe mempty $ lookup "id" blockAttrKeyValues
+    classes = case lookup "class" blockAttrKeyValues of
+                Nothing     -> []
+                Just clsStr -> words clsStr
+    kv      = filter ((`notElem` ["id", "class"]) . fst) blockAttrKeyValues
+  in (ident, classes, kv)
+
 stringyMetaAttribute :: (String -> Bool) -> OrgParser (String, String)
 stringyMetaAttribute attrCheck = try $ do
   metaLineStart
@@ -282,9 +322,9 @@ blockAttributes = try $ do
   let kvAttrs = foldl' (appendValues "ATTR_HTML") Nothing kv
   let name    = lookup "NAME" kv
   let label   = lookup "LABEL" kv
-  caption' <- maybe (return Nothing)
-                    (fmap Just . parseFromString inlines)
-                    caption
+  caption' <- case caption of
+                   Nothing -> return Nothing
+                   Just s  -> Just <$> parseFromString inlines (s ++ "\n")
   kvAttrs' <- parseFromString keyValues . (++ "\n") $ fromMaybe mempty kvAttrs
   return $ BlockAttributes
            { blockAttrName = name
@@ -336,23 +376,25 @@ orgBlock = try $ do
   blockAttrs <- blockAttributes
   blkType <- blockHeaderStart
   ($ blkType) $
-    case blkType of
+    case (map toLower blkType) of
       "export"  -> exportBlock
       "comment" -> rawBlockLines (const mempty)
-      "html"    -> rawBlockLines (return . (B.rawBlock blkType))
-      "latex"   -> rawBlockLines (return . (B.rawBlock blkType))
-      "ascii"   -> rawBlockLines (return . (B.rawBlock blkType))
+      "html"    -> rawBlockLines (return . B.rawBlock (lowercase blkType))
+      "latex"   -> rawBlockLines (return . B.rawBlock (lowercase blkType))
+      "ascii"   -> rawBlockLines (return . B.rawBlock (lowercase blkType))
       "example" -> rawBlockLines (return . exampleCode)
       "quote"   -> parseBlockLines (fmap B.blockQuote)
       "verse"   -> verseBlock
       "src"     -> codeBlock blockAttrs
-      _         -> parseBlockLines (fmap $ B.divWith (mempty, [blkType], mempty))
+      _         -> parseBlockLines $
+                   let (ident, classes, kv) = attrFromBlockAttributes blockAttrs
+                   in fmap $ B.divWith (ident, classes ++ [blkType], kv)
  where
    blockHeaderStart :: OrgParser String
-   blockHeaderStart = try $ do
-     skipSpaces
-     blockType <- stringAnyCase "#+begin_" *> orgArgWord
-     return (map toLower blockType)
+   blockHeaderStart = try $ skipSpaces *> stringAnyCase "#+begin_" *> orgArgWord
+
+   lowercase :: String -> String
+   lowercase = map toLower
 
 rawBlockLines :: (String   -> F Blocks) -> String -> OrgParser (F Blocks)
 rawBlockLines f blockType = (ignHeaders *> (f <$> rawBlockContent blockType))
@@ -421,8 +463,19 @@ verseBlock :: String -> OrgParser (F Blocks)
 verseBlock blockType = try $ do
   ignHeaders
   content <- rawBlockContent blockType
-  fmap B.para . mconcat . intersperse (pure B.linebreak)
-    <$> mapM (parseFromString inlines) (map (++ "\n") . lines $ content)
+  fmap B.lineBlock . sequence
+    <$> mapM parseVerseLine (lines content)
+ where
+   -- replace initial spaces with nonbreaking spaces to preserve
+   -- indentation, parse the rest as normal inline
+   parseVerseLine :: String -> OrgParser (F Inlines)
+   parseVerseLine cs = do
+     let (initialSpaces, indentedLine) = span isSpace cs
+     let nbspIndent = if null initialSpaces
+                      then mempty
+                      else B.str $ map (const '\160') initialSpaces
+     line <- parseFromString inlines (indentedLine ++ "\n")
+     return (trimInlinesF $ pure nbspIndent <> line)
 
 -- | Read a code block and the associated results block if present.  Which of
 -- boths blocks is included in the output is determined using the "exports"
@@ -566,28 +619,44 @@ propertiesDrawer = try $ do
 -- Figures
 --
 
--- | Figures (Image on a line by itself, preceded by name and/or caption)
+-- | Figures or an image paragraph (i.e. an image on a line by itself). Only
+-- images with a caption attribute are interpreted as figures.
 figure :: OrgParser (F Blocks)
 figure = try $ do
   figAttrs <- blockAttributes
-  src <- skipSpaces *> selfTarget <* skipSpaces <* newline
-  guard . not . isNothing . blockAttrCaption $ figAttrs
-  guard (isImageFilename src)
-  let figName    = fromMaybe mempty $ blockAttrName figAttrs
-  let figLabel   = fromMaybe mempty $ blockAttrLabel figAttrs
-  let figCaption = fromMaybe mempty $ blockAttrCaption figAttrs
-  let figKeyVals = blockAttrKeyValues figAttrs
-  let attr       = (figLabel, mempty, figKeyVals)
-  return $ (B.para . B.imageWith attr src (withFigPrefix figName) <$> figCaption)
+  src <- skipSpaces *> selfTarget <* skipSpaces <* endOfParagraph
+  case cleanLinkString src of
+    Nothing     -> mzero
+    Just imgSrc -> do
+      guard (isImageFilename imgSrc)
+      let isFigure = not . isNothing $ blockAttrCaption figAttrs
+      return $ imageBlock isFigure figAttrs imgSrc
  where
+   selfTarget :: OrgParser String
+   selfTarget = try $ char '[' *> linkTarget <* char ']'
+
+   imageBlock :: Bool -> BlockAttributes -> String -> F Blocks
+   imageBlock isFigure figAttrs imgSrc =
+     let
+       figName    = fromMaybe mempty $ blockAttrName figAttrs
+       figLabel   = fromMaybe mempty $ blockAttrLabel figAttrs
+       figCaption = fromMaybe mempty $ blockAttrCaption figAttrs
+       figKeyVals = blockAttrKeyValues figAttrs
+       attr       = (figLabel, mempty, figKeyVals)
+       figTitle   = (if isFigure then withFigPrefix else id) figName
+     in
+       B.para . B.imageWith attr imgSrc figTitle <$> figCaption
+
    withFigPrefix :: String -> String
    withFigPrefix cs =
      if "fig:" `isPrefixOf` cs
      then cs
      else "fig:" ++ cs
 
-   selfTarget :: OrgParser String
-   selfTarget = try $ char '[' *> linkTarget <* char ']'
+-- | Succeeds if looking at the end of the current paragraph
+endOfParagraph :: OrgParser ()
+endOfParagraph = try $ skipSpaces *> newline *> endOfBlock
+
 
 --
 -- Examples
@@ -610,85 +679,41 @@ exampleCode = B.codeBlockWith ("", ["example"], [])
 --
 
 specialLine :: OrgParser (F Blocks)
-specialLine = fmap return . try $ metaLine <|> commentLine
+specialLine = fmap return . try $ rawExportLine <|> metaLine <|> commentLine
 
--- The order, in which blocks are tried, makes sure that we're not looking at
--- the beginning of a block, so we don't need to check for it
-metaLine :: OrgParser Blocks
-metaLine = mempty <$ metaLineStart <* (optionLine <|> declarationLine)
+rawExportLine :: OrgParser Blocks
+rawExportLine = try $ do
+  metaLineStart
+  key <- metaKey
+  if key `elem` ["latex", "html", "texinfo", "beamer"]
+    then B.rawBlock key <$> anyLine
+    else mzero
 
 commentLine :: OrgParser Blocks
 commentLine = commentLineStart *> anyLine *> pure mempty
 
-declarationLine :: OrgParser ()
-declarationLine = try $ do
-  key   <- metaKey
-  value <- metaInlines
-  updateState $ \st ->
-    let meta' = B.setMeta key <$> value <*> pure nullMeta
-    in st { orgStateMeta = orgStateMeta st <> meta' }
-
-metaInlines :: OrgParser (F MetaValue)
-metaInlines = fmap (MetaInlines . B.toList) <$> inlinesTillNewline
-
-metaKey :: OrgParser String
-metaKey = map toLower <$> many1 (noneOf ": \n\r")
-                      <*  char ':'
-                      <*  skipSpaces
-
-optionLine :: OrgParser ()
-optionLine = try $ do
-  key <- metaKey
-  case key of
-    "link"    -> parseLinkFormat >>= uncurry addLinkFormat
-    "options" -> exportSettings
-    _         -> mzero
-
-addLinkFormat :: String
-              -> (String -> String)
-              -> OrgParser ()
-addLinkFormat key formatter = updateState $ \s ->
-  let fs = orgStateLinkFormatters s
-  in s{ orgStateLinkFormatters = M.insert key formatter fs }
-
-parseLinkFormat :: OrgParser ((String, String -> String))
-parseLinkFormat = try $ do
-  linkType <- (:) <$> letter <*> many (alphaNum <|> oneOf "-_") <* skipSpaces
-  linkSubst <- parseFormat
-  return (linkType, linkSubst)
-
--- | An ad-hoc, single-argument-only implementation of a printf-style format
--- parser.
-parseFormat :: OrgParser (String -> String)
-parseFormat = try $ do
-  replacePlain <|> replaceUrl <|> justAppend
- where
-   -- inefficient, but who cares
-   replacePlain = try $ (\x -> concat . flip intersperse x)
-                     <$> sequence [tillSpecifier 's', rest]
-   replaceUrl   = try $ (\x -> concat . flip intersperse x . urlEncode)
-                     <$> sequence [tillSpecifier 'h', rest]
-   justAppend   = try $ (++) <$> rest
-
-   rest            = manyTill anyChar         (eof <|> () <$ oneOf "\n\r")
-   tillSpecifier c = manyTill (noneOf "\n\r") (try $ string ('%':c:""))
-
 
 --
 -- Tables
 --
+data ColumnProperty = ColumnProperty
+  { columnAlignment :: Maybe Alignment
+  , columnRelWidth  :: Maybe Int
+  } deriving (Show, Eq)
+
+instance Default ColumnProperty where
+  def = ColumnProperty Nothing Nothing
 
 data OrgTableRow = OrgContentRow (F [Blocks])
-                 | OrgAlignRow [Alignment]
+                 | OrgAlignRow [ColumnProperty]
                  | OrgHlineRow
 
 -- OrgTable is strongly related to the pandoc table ADT.  Using the same
 -- (i.e. pandoc-global) ADT would mean that the reader would break if the
 -- global structure was to be changed, which would be bad.  The final table
--- should be generated using a builder function.  Column widths aren't
--- implemented yet, so they are not tracked here.
+-- should be generated using a builder function.
 data OrgTable = OrgTable
-  { orgTableAlignments :: [Alignment]
+  { orgTableColumnProperties :: [ColumnProperty]
   , orgTableHeader     :: [Blocks]
   , orgTableRows       :: [[Blocks]]
   }
@@ -705,8 +730,20 @@ table = try $ do
 orgToPandocTable :: OrgTable
                  -> Inlines
                  -> Blocks
-orgToPandocTable (OrgTable aligns heads lns) caption =
-  B.table caption (zip aligns $ repeat 0) heads lns
+orgToPandocTable (OrgTable colProps heads lns) caption =
+  let totalWidth = if any (not . isNothing) (map columnRelWidth colProps)
+                   then Just . sum $ map (fromMaybe 1 . columnRelWidth) colProps
+                   else Nothing
+  in B.table caption (map (convertColProp totalWidth) colProps) heads lns
+ where
+   convertColProp :: Maybe Int -> ColumnProperty -> (Alignment, Double)
+   convertColProp totalWidth colProp =
+     let
+       align' = fromMaybe AlignDefault $ columnAlignment colProp
+       width' = fromMaybe 0 $ (\w t -> (fromIntegral w / fromIntegral t))
+                              <$> (columnRelWidth colProp)
+                              <*> totalWidth
+     in (align', width')
 
 tableRows :: OrgParser [OrgTableRow]
 tableRows = try $ many (tableAlignRow <|> tableHline <|> tableContentRow)
@@ -722,22 +759,22 @@ tableContentCell = try $
 tableAlignRow :: OrgParser OrgTableRow
 tableAlignRow = try $ do
   tableStart
-  cells <- many1Till tableAlignCell newline
+  colProps <- many1Till columnPropertyCell newline
   -- Empty rows are regular (i.e. content) rows, not alignment rows.
-  guard $ any (/= AlignDefault) cells
-  return $ OrgAlignRow cells
+  guard $ any (/= def) colProps
+  return $ OrgAlignRow colProps
 
-tableAlignCell :: OrgParser Alignment
-tableAlignCell =
-  choice [ try $ emptyCell *> return AlignDefault
-         , try $ skipSpaces
-                   *> char '<'
-                   *> tableAlignFromChar
-                   <* many digit
-                   <* char '>'
-                   <* emptyCell
-         ] <?> "alignment info"
-    where emptyCell = try $ skipSpaces *> endOfCell
+columnPropertyCell :: OrgParser ColumnProperty
+columnPropertyCell = emptyCell <|> propCell <?> "alignment info"
+ where
+   emptyCell = ColumnProperty Nothing Nothing <$ (try $ skipSpaces *> endOfCell)
+   propCell = try $ ColumnProperty
+                 <$> (skipSpaces
+                      *> char '<'
+                      *> optionMaybe tableAlignFromChar)
+                 <*> (optionMaybe (many1 digit >>= safeRead)
+                      <* char '>'
+                      <* emptyCell)
 
 tableAlignFromChar :: OrgParser Alignment
 tableAlignFromChar = try $
@@ -759,7 +796,8 @@ rowsToTable = foldM rowToContent emptyTable
  where emptyTable = OrgTable mempty mempty mempty
 
 normalizeTable :: OrgTable -> OrgTable
-normalizeTable (OrgTable aligns heads rows) = OrgTable aligns' heads rows
+normalizeTable (OrgTable colProps heads rows) =
+  OrgTable colProps' heads rows
  where
    refRow = if heads /= mempty
             then heads
@@ -768,7 +806,7 @@ normalizeTable (OrgTable aligns heads rows) = OrgTable aligns' heads rows
                    _     -> mempty
    cols = length refRow
    fillColumns base padding = take cols $ base ++ repeat padding
-   aligns' = fillColumns aligns AlignDefault
+   colProps' = fillColumns colProps def
 
 -- One or more horizontal rules after the first content line mark the previous
 -- line as a header.  All other horizontal lines are discarded.
@@ -778,7 +816,7 @@ rowToContent :: OrgTable
 rowToContent orgTable row =
   case row of
     OrgHlineRow       -> return singleRowPromotedToHeader
-    OrgAlignRow as    -> return . setAligns $ as
+    OrgAlignRow props -> return . setProperties $ props
     OrgContentRow cs  -> appendToBody cs
  where
    singleRowPromotedToHeader :: OrgTable
@@ -787,8 +825,8 @@ rowToContent orgTable row =
             orgTable{ orgTableHeader = b , orgTableRows = [] }
      _   -> orgTable
 
-   setAligns :: [Alignment] -> OrgTable
-   setAligns aligns = orgTable{ orgTableAlignments = aligns }
+   setProperties :: [ColumnProperty] -> OrgTable
+   setProperties ps = orgTable{ orgTableColumnProperties = ps }
 
    appendToBody :: F [Blocks] -> F OrgTable
    appendToBody frow = do
@@ -849,9 +887,6 @@ paraOrPlain = try $ do
        *> return (B.para <$> ils))
     <|>  (return (B.plain <$> ils))
 
-inlinesTillNewline :: OrgParser (F Inlines)
-inlinesTillNewline = trimInlinesF . mconcat <$> manyTill inline newline
-
 
 --
 -- list blocks
